import functools
import pickle
import re
import threading
import time

import numpy as np
from karabo.bound import (
    FLOAT_ELEMENT,
    IMAGEDATA_ELEMENT,
    INPUT_CHANNEL,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    ChannelMetaData,
    Dims,
    Encoding,
    Epochstamp,
    Hash,
    ImageData,
    MetricPrefix,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES
from TrainMatcher import TrainMatcher
from TrainMatcher import scenes as trainmatcher_scenes

from . import scenes
from ._version import version as deviceVersion

preview_schema = Schema()
(
    IMAGEDATA_ELEMENT(preview_schema).key("image").commit(),

    UINT64_ELEMENT(preview_schema).key("trainId").readOnly().commit(),
)

xtdf_source_re = re.compile(r".*\/DET\/(\d+)CH0:xtdf")
daq_source_re = re.compile(r".*\/DET\/.*?(\d+):daqOutput")


# TODO: merge scene with TrainMatcher's nice overview
@KARABO_CLASSINFO("SimpleAssembler", deviceVersion)
class SimpleAssembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["scene", "assemblerOverview"])
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("processingTime")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(500)
            .info("Cannot keep up with GUI limit")
            .needsAcknowledging(False)
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("timeOfFlight")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(1000)
            .info("Time of flight exceeding 1 s")
            .needsAcknowledging(False)
            .commit(),

            STRING_ELEMENT(expected)
            .key("pathToStack")
            .assignmentOptional()
            .defaultValue("image.data")
            .commit(),

            UINT32_ELEMENT(expected)
            .key("downsamplingFactor")
            .description(
                "If greater than 1, the assembled image will be downsampled by this "
                "factor in x and y dimensions before sending. This is only to save "
                "bandwidth in case GUI updates start lagging."
            )
            .assignmentOptional()
            .defaultValue(1)
            .options("1,2,4,8")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("downsamplingFunction")
            .description("Reduction function used during downsampling.")
            .assignmentOptional()
            .defaultValue("nanmax")
            .options("nanmax,nanmean,nanmin,nanmedian")
            .reconfigurable()
            .commit(),

            INPUT_CHANNEL(expected)
            .key("geometryInput")
            .displayedName("Geometry input")
            .commit(),

            OUTPUT_CHANNEL(expected)  # can OVERWRITE_ELEMENT even do this?
            .key("output")
            .dataSchema(preview_schema)
            .commit(),
        )

    def initialization(self):
        super().initialization()

        # TODO: match inside device, fill multiple independent buffers

        self._path_to_stack = self.get("pathToStack")
        self.geometry = None
        self.input_buffer = None

        self.KARABO_ON_DATA("geometryInput", self.receive_geometry)
        self.KARABO_SLOT(self.requestScene)

        self.ask_for_geometry()
        self.start()

    def requestScene(self, params):
        # TODO: unify with TrainMatcher overview
        scene_name = params.get("name", default="")
        if scene_name == "assemblerOverview":
            payload = Hash("name", scene_name, "success", True)
            payload["data"] = scenes.simple_assembler_overview(
                device_id=self.getInstanceId(),
                geometry_device_id=self.get("geometryInput.connectedOutputChannels")[
                    0
                ].split(":")[0],
            )
            self.reply(
                Hash(
                    "type",
                    "deviceScene",
                    "origin",
                    self.getInstanceId(),
                    "payload",
                    payload,
                )
            )
        else:
            return super().requestScene(params)

    def receive_geometry(self, data, metadata):
        self.log.INFO("Received a new geometry")
        self.geometry = pickle.loads(data.get("pickledGeometry"))
        # TODO: allow multiple memory cells (extra geom notion of extra dimensions)
        self.input_buffer = np.zeros(self.geometry.expected_data_shape)

    def ask_for_geometry(self):
        def runner():
            self.log.INFO("Will ask around for a geometry")
            max_tries = 10
            for i in range(max_tries):
                time.sleep(np.random.random() * 10)
                if self.geometry is None:
                    missing_connections = set(
                        self.get("geometryInput.missingConnections")
                    )
                    # note: connectedOutputChannels not necessarily connected...
                    geometry_device_list = [
                        channel
                        for channel in self.get("geometryInput.connectedOutputChannels")
                        if channel not in missing_connections
                    ]
                    if not geometry_device_list:
                        self.log.INFO("No geometry device connected")
                        continue
                    geometry_device = geometry_device_list[0].split(":")[0]
                    self.log.INFO(f"Asking {geometry_device} for a geometry")
                    self.signalSlotable.call(geometry_device, "pleaseSendYourGeometry")
                    time.sleep(1)

                if self.geometry is not None:
                    return
            self.log.INFO(f"Failed to get geometry in {max_tries} tries, need help")
        threading.Thread(target=runner, daemon=True).start()

    def _send(self, train_id, sources):
        # TODO: adapt to appropriate hook for new TrainMatcher (no _send)
        if self.geometry is None:
            self.log.WARN("Have not received a geometry yet")
            return

        timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))

        module_indices_unfilled = set(range(self.input_buffer.shape[0]))
        for source, (data, metadata) in sources.items():
            # TODO: handle failure to "parse" source, get data out
            module_index = self._source_to_index(source)
            self.input_buffer[module_index] = np.squeeze(data.get(self._path_to_stack))
            module_indices_unfilled.discard(module_index)

        for unfilled_module in module_indices_unfilled:
            self.input_buffer[unfilled_module].fill(0)
            # TODO: configurable treatment of missing modules

        # TODO: reusable output buffer to save on allocation
        assembled, _ = self.geometry.position_modules_fast(self.input_buffer)

        downsampling_factor = self.get("downsamplingFactor")
        if downsampling_factor > 1:
            assembled = downsample_2d(
                assembled,
                downsampling_factor,
                reduction_fun=getattr(np, self.get("downsamplingFunction"))
            )

        # TODO: optionally include control data
        out_hash = Hash(
            "image",
            ImageData(
                # TODO: get around this being mirrored...
                (assembled[::-1, ::-1]).astype(np.int32),
                Dims(*assembled.shape),
                Encoding.GRAY,
            ),
            "trainId",
            train_id,
        )
        channel = self.signalSlotable.getOutputChannel("output")
        channel.write(out_hash, ChannelMetaData(self.getInstanceId(), timestamp))
        channel.update()
        self.rate_out.update()

    def _update_rate(self):
        self._buffered_status_update.set(
            "processingTime", self._processing_time_ema.get()
        )
        self._buffered_status_update.set("timeOfFlight", self._time_of_flight_ema.get())
        self._buffered_status_update.set("rate", self._rate_tracker.get())
        self.set(self._buffered_status_update)

    @functools.lru_cache()
    def _source_to_index(self, source):
        # note: cache means warning only shows up once (also not performance-critical)
        # TODO: allow user to inspect, modify the mapping

        match = xtdf_source_re.match(source)
        if match is not None:
            return int(match.group(1))

        match = daq_source_re.match(source)
        if match is not None:
            return int(match.group(1)) - 1

        self.log.WARN(f"Couldn't figure out index for source {source}")
        return 0


def downsample_2d(arr, factor, reduction_fun=np.nanmax):
    """Generalization of downsampling from FemDataAssembler

    Expects first two dimensions of arr to be multiple of 2 ** factor
    Useful if you're sitting at home and ssh connection is slow to get full-resolution
    previews."""

    for i in range(factor // 2):
        arr = reduction_fun(
            (
                arr[:-1:2],
                arr[1::2],
            ), axis=0
        )
        arr = reduction_fun(
            (
                arr[:, :-1:2],
                arr[:, 1::2],
            ), axis=0
        )
    return arr
